#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ranlib.h"
#include "util.h"
#include "blackScholes.h"


double ceuropean(int underliers, double *price, double strike, double interest, double variance, double dividend,
		double tolerance){

  double *parm, *scratch, *dam, *covariance, *growth, error, *newprice,f,g,beta,G,
    answer,count=0.0,w;
  double S1=0.0,S2=0.0,S3=0.0,S4=0.0,S5=0.0,S6=0.0,S7=0.0,S8=0.0,S9=0.0,S10=0.0,S11=0.0,S12=0.0,S13=0.0,S14=0.0,
    S15=0.0,S16=0.0,S17=0.0;
  int i,j,k,max1, max2;

  
  allocate1dDouble(&parm, underliers*(underliers+3)/2+1);
  allocate1dDouble(&scratch, underliers);
  allocate1dDouble(&dam, underliers);
  allocate1dDouble(&covariance, underliers*underliers);
  allocate1dDouble(&growth, underliers);
  allocate1dDouble(&newprice, underliers);

  if(price[0] > price[1]){
    max1 = 0;
    max2 = 1;
  }else{
    max1 = 1;
    max2 = 0;
  }

  for(i=2; i<underliers; ++i){
    if(price[i] > price[max1]){
      max2 = max1;
      max1 = i;
    }else if(price[i] > price[max2]){
      max2 = i;
    }
  }
  
  G = maxOf2Risk(price[max1],
				price[max2],
				strike,
				1.0,
				sqrt(variance),
				sqrt(variance),
				interest,
				dividend,
				dividend,
				0.0);

  k = 0;
  for(i=0; i<underliers; ++i){
    for(j=0; j<underliers; ++j){
      if(i==j){
	covariance[k] = variance;
	dam[i] = interest - dividend - variance/2.0;
      }else{
	covariance[k] = 0;
      }
      k++;
    }
  }

  setgmn(dam, covariance, underliers, parm);

  do{
    for(i=0; i<100; ++i){
      genmn(parm, growth, scratch);

      newprice[0] = price[0]*exp(growth[0]);
      f = newprice[0];
      for(j=1; j<underliers; j++){
	newprice[j] = price[j]*exp(growth[j]);
	if(newprice[j] > f) f = newprice[j];
      }
      f = dMax(0, exp(-interest)*(f - strike));
      
      g = dMax(newprice[max1],newprice[max2]);
      g = dMax(0, exp(-interest)*(g - strike));

      count += 1.0;

      w = 1.0;
      S1 += w; 
      S2 += w*w; 
      S7 += w*f; 
      S8 += w*(g-G); 
      S9 += w*w*f; 
      S10 += w*w*(g-G); 
      S11 += w*w*f*f; 
      S12 += w*w*f*(g-G); 
      S13 += w*w*(g-G)*(g-G); 
      S14 += w*f*(g-G); 
      S15 += w*(g-G)*(g-G); 
      S16 += w*f*f; 
      S17 += w * g; 
    }
    beta = (S12/S1 - (S8*S9 + S7*S10)/(S1*S1) + S2*S7*S8/(S1*S1*S1))/
      (S13/S1 - 2*S8*S10/(S1*S1) + S2*S8*S8/(S1*S1*S1));
    
    answer = (S7-beta*S8)/S1;
    
    error = sqrt(1/(S1*S1)*((S11 -2*beta*S12 + beta*beta*S13)   
			-(2/S1)*(S7-beta*S8)*(S9-beta*S10)  
			+(S2/(S1*S1))*(S7*S7 -2*beta*S7*S8 +beta*beta*S8*S8) )); 

  }while(fabs(error/answer) > tolerance);

  deallocate1dDouble(&parm, underliers*(underliers+3)/2+1);
  deallocate1dDouble(&scratch, underliers);
  deallocate1dDouble(&dam, underliers);
  deallocate1dDouble(&covariance, underliers*underliers);
  deallocate1dDouble(&growth, underliers);
  deallocate1dDouble(&newprice, underliers);
  
  return(answer);
}


double european(int underliers, double *price, double strike, double interest, double variance, double dividend,
		double tolerance){

  double *parm, *scratch, *dam, *covariance, *growth, S1=0.0,S2=0.0,S3=0.0,S4=0.0,S5=0.0, error, *newprice,f,g,beta,G,
    answer,count=0.0;
  int i,j,k,max1, max2;

  allocate1dDouble(&parm, underliers*(underliers+3)/2+1);
  allocate1dDouble(&scratch, underliers);
  allocate1dDouble(&dam, underliers);
  allocate1dDouble(&covariance, underliers*underliers);
  allocate1dDouble(&growth, underliers);
  allocate1dDouble(&newprice, underliers);

  k = 0;
  for(i=0; i<underliers; ++i){
    for(j=0; j<underliers; ++j){
      if(i==j){
	covariance[k] = variance;
	dam[i] = interest - dividend - variance/2.0;
      }else{
	covariance[k] = 0;
      }
      k++;
    }
  }

  setgmn(dam, covariance, underliers, parm);

  do{
    for(i=0; i<100; ++i){
      genmn(parm, growth, scratch);

      f = newprice[0] = price[0]*exp(growth[0]);
      for(j=1; j<underliers; j++){
	newprice[j] = price[j]*exp(growth[j]);
	if(newprice[j] > f) f = newprice[j];
      }
      f = dMax(0, exp(-interest)*(f - strike));
      
      count += 1.0;

      S1 += f;
      S2 += f*f;
    }

    answer = S1/count;
    error = sqrt((S2/count - S1/count*S1/count)/count);
    
  }while(fabs(error/answer) > tolerance);

  deallocate1dDouble(&parm, underliers*(underliers+3)/2+1);
  deallocate1dDouble(&scratch, underliers);
  deallocate1dDouble(&dam, underliers);
  deallocate1dDouble(&covariance, underliers*underliers);
  deallocate1dDouble(&growth, underliers);
  deallocate1dDouble(&newprice, underliers);

  return(answer);
}
